import abc

class Source(metaclass=abc.ABCMeta):

  @abc.abstractmethod
  def wait_for_event(self):
    pass

  @abc.abstractmethod
  def start(self):
    pass

  @abc.abstractmethod
  def set_source_wrapper(self, source_wrapper):
    pass

  @abc.abstractmethod
  def event_success(self, exchange):
      pass

  # todo exchange with error
  @abc.abstractmethod
  def event_failure(self, err, exchange):
    pass



from .eipiphany_context_termination import EipiphanyContextTermination


class DefaultEipiphanyContextTermination(EipiphanyContextTermination):
  def is_terminate(self, context):
    if not len(context.processes):
      return True
    for process in context.processes:
      if not process.is_alive():
        return True
    return False



import time

from ..base.default_eipiphany_context_termination import DefaultEipiphanyContextTermination


class TestEipiphanyContextTermination(DefaultEipiphanyContextTermination):
  def __init__(self, timeout=20000):
    self.__timeout = timeout
    self.__mock_endpoints = {}

  @property
  def mock_endpoints(self):
    return self.__mock_endpoints

  @mock_endpoints.setter
  def mock_endpoints(self, value):
    self.__mock_endpoints = value

  def is_terminate(self, context):
    term = super().is_terminate(context)
    if term:
      return True
    if self.__mock_endpoints.values():
      completed = 0
      for ep in self.__mock_endpoints.values():
        if len(ep.exchanges) >= ep.expected_message_count:
          completed += 1
      if completed == len(self.__mock_endpoints.values()):
        return True
    current_time = round(time.time() * 1000)
    return current_time - context.start_time >= self.__timeout



import time
from multiprocessing import Manager

from .default_eipiphany_context_termination import DefaultEipiphanyContextTermination


class EipiphanyContext(object):
  def __init__(self, termination=DefaultEipiphanyContextTermination()):
    self.__manager = Manager()
    self._routes = []
    self.__processes = []
    self.__start_time = None
    self._termination = termination

  def add_route_builder(self, route_builder):
    route_builder.build()
    for route in route_builder.get_routes():
      self._routes.append(route)
    return self

  def __terminate(self):
    for process in self.__processes:
      process.terminate()
      process.join()
      process.close()

  def start(self):
    self.__start_time = round(time.time() * 1000)
    for route in self._routes:
      for process in route.start():
        self.__processes.append(process)
    terminate = False
    while not terminate:
      time.sleep(1)
      terminate = self._termination.is_terminate(self)
    self.__terminate()

  @property
  def manager(self):
    return self.__manager

  @property
  def processes(self):
    return self.__processes

  @property
  def start_time(self):
    return self.__start_time

  def __enter__(self):
    self.__manager.__enter__()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.__manager.__exit__(exc_type, exc_val, exc_tb)





class ExchangeHandler(object):
  def __init__(self):
    self.__processor = None
    self.__filter = None

  @property
  def processor(self):
    return self.__processor

  @processor.setter
  def processor(self, value):
    self.set_processor(value)

  def set_processor(self, value):
    self.__processor = value
    return self

  @property
  def filter(self):
    return self.__filter

  @filter.setter
  def filter(self, value):
    self.set_filter(value)

  def set_filter(self, value):
    self.__filter = value
    return self



import logging

from ...framework.base.error_handler import ErrorHandler

logger = logging.getLogger(__name__)


class DefaultErrorHandler(ErrorHandler):

  def handle_exception(self, exchange):
    exception = exchange.get_header(ErrorHandler.EXCEPTION_CAUGHT)
    logger.error("Exception in route", exc_info=exception)




from .mock_endpoint import MockEndpoint
from .test_eipiphany_context_termination import TestEipiphanyContextTermination
from ..base.eipiphany_context import EipiphanyContext
from ..internal.exchange_handler import ExchangeHandler


class EipiphanyTestContext(EipiphanyContext):
  def __init__(self, original_context,
      termination=TestEipiphanyContextTermination()):
    self.__original_context = original_context
    self.__mock_endpoints = {}
    self.__original_context._termination = termination
    self.__original_context._termination.mock_endpoints = self.__mock_endpoints

  def add_route_builder(self, route_builder):
    self.__original_context.add_route_builder(route_builder)
    return self

  def start(self):
    self.__original_context.start()

  @property
  def manager(self):
    return self.__original_context.manager

  @property
  def processes(self):
    return self.__original_context.processes

  @property
  def start_time(self):
    return self.__original_context.start_time

  def __enter__(self):
    self.__original_context.__enter__()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.__original_context.__exit__(exc_type, exc_val, exc_tb)

  def __find_route(self, route_id):
    the_route = None
    for route in self.__original_context._routes:
      if route.route_id == route_id:
        the_route = route
    if not the_route:
      raise Exception("Route with id " + route_id + " was not found")
    return the_route

  def mock_endpoint(self, route_id, existing_processor,
      expected_message_count=1):
    the_route = self.__find_route(route_id)
    found = []
    exchange_handlers = the_route._exchange_handlers
    for i in range(len(exchange_handlers)):
      if existing_processor == exchange_handlers[i].processor:
        found.append(i)
    if found:
      offset = 0
      mock_endpoint = MockEndpoint(self.__original_context,
                                   expected_message_count=expected_message_count)
      for i in found:
        exchange_handlers.insert(i + 1 + offset,
                                 ExchangeHandler().set_processor(mock_endpoint))
        offset = offset + 1
      self.__mock_endpoints[existing_processor] = mock_endpoint
    return self

  def mock_endpoint_and_skip(self, route_id, existing_processor,
      expected_message_count=1):
    the_route = self.__find_route(route_id)
    found = []
    exchange_handlers = the_route._exchange_handlers
    for i in range(len(exchange_handlers)):
      if existing_processor == exchange_handlers[i].processor:
        found.append(i)
    if found:
      mock_endpoint = MockEndpoint(self.__original_context,
                                   expected_message_count=expected_message_count)
      for i in found:
        exchange_handlers[i] = ExchangeHandler().set_processor(mock_endpoint)
      self.__mock_endpoints[existing_processor] = mock_endpoint
    return self

  def insert_before_processor(self, route_id, existing_processor, processor):
    the_route = self.__find_route(route_id)
    found = []
    exchange_handlers = the_route._exchange_handlers
    for i in range(len(exchange_handlers)):
      if existing_processor == exchange_handlers[i].processor:
        found.append(i)
    if found:
      offset = 0
      for i in found:
        exchange_handlers.insert(i + offset,
                                 ExchangeHandler().set_processor(processor))
        offset = offset + 1
    return self

  def insert_after_processor(self, route_id, existing_processor, processor):
    the_route = self.__find_route(route_id)
    found = []
    exchange_handlers = the_route._exchange_handlers
    for i in range(len(exchange_handlers)):
      if existing_processor == exchange_handlers[i].processor:
        found.append(i)
    if found:
      offset = 0
      for i in found:
        exchange_handlers.insert(i + 1 + offset,
                                 ExchangeHandler().set_processor(processor))
        offset = offset + 1
    return self

  def replace_processor(self, route_id, existing_processor, processor):
    the_route = self.__find_route(route_id)
    found = []
    exchange_handlers = the_route._exchange_handlers
    for i in range(len(exchange_handlers)):
      if existing_processor == exchange_handlers[i].processor:
        found.append(i)
    if found:
      for i in found:
        exchange_handlers[i] = ExchangeHandler().set_processor(processor)
    return self

  def remove_processor(self, route_id, existing_processor):
    the_route = self.__find_route(route_id)
    found = []
    exchange_handlers = the_route._exchange_handlers
    for i in range(len(exchange_handlers)):
      if existing_processor == exchange_handlers[i].processor:
        found.append(i)
    offset = 0
    if found:
      for i in found:
        exchange_handlers.pop(i + offset)
        offset = offset - 1
    return self

  def replace_from_with(self, route_id, source):
    the_route = self.__find_route(route_id)
    the_route._source_wrapper._source = source
    return self

  @property
  def mock_endpoints(self):
    return self.__mock_endpoints



from ..base.processor import Processor


class MockEndpoint(Processor):
  def __init__(self, context, expected_message_count):
    super().__init__()
    self.__exchanges = context.manager.list()
    self.__expected_message_count = expected_message_count

  def process(self, exchange):
    self.exchanges.append(exchange)

  @property
  def exchanges(self):
    return self.__exchanges

  @property
  def expected_message_count(self):
    return self.__expected_message_count